@@ -245,6 +245,14 @@ class Test_FMUModelCS1:
         assert os.path.exists("Test_log.txt")
 
 class Test_FMUModelBase:
+    @testattr(stddist = True)
+    def test_unicode_description(self):
+        full_path = os.path.join(file_path, "files", "FMUs", "XML", "ME1.0", "Description.fmu")
+        model = FMUModelME1(full_path, _connect_dll=False)
+        
+        desc = model.get_variable_description("x")
+
+        assert desc == "Test symbols '' ‘’"
     
     @testattr(stddist = True)
     def test_get_erronous_nominals(self):
@@ -327,6 +335,92 @@ class Test_FMUModelCS2:
 
 if assimulo_installed:
     class Test_FMUModelME2_Simulation:
+        @testattr(stddist = True)
+        def test_basicsens1(self):
+            #Noncompliant FMI test as 'd' is parameter is not supposed to be able to be set during simulation
+            model = Dummy_FMUModelME2([], "BasicSens1.fmu", os.path.join(file_path, "files", "FMUs", "XML", "ME2.0"), _connect_dll=False)
+            
+            def f(*args, **kwargs):
+                d = model.values[model.variables["d"].value_reference]
+                x = model.continuous_states[0]
+                model.values[model.variables["der(x)"].value_reference] = d*x
+                return np.array([d*x])
+            
+            model.get_derivatives = f
+            
+            opts = model.simulate_options()
+            opts["sensitivities"] = ["d"]
+
+            res = model.simulate(options=opts)
+            nose.tools.assert_almost_equal(res.final('dx/dd'), 0.36789, 3)
+            
+            assert res.solver.statistics["nsensfcnfcns"] > 0
+        
+        @testattr(stddist = True)
+        def test_basicsens1dir(self):
+            model = Dummy_FMUModelME2([], "BasicSens1.fmu", os.path.join(file_path, "files", "FMUs", "XML", "ME2.0"), _connect_dll=False)
+            
+            caps = model.get_capability_flags()
+            caps["providesDirectionalDerivatives"] = True
+            model.get_capability_flags = lambda : caps
+            
+            def f(*args, **kwargs):
+                d = model.values[model.variables["d"].value_reference]
+                x = model.continuous_states[0]
+                model.values[model.variables["der(x)"].value_reference] = d*x
+                return np.array([d*x])
+            
+            def d(*args, **kwargs):
+                if args[0][0] == 40:
+                    return np.array([-1.0])
+                else:
+                    return model.continuous_states
+                
+            model.get_directional_derivative = d
+            model.get_derivatives = f
+            model._provides_directional_derivatives = lambda : True
+            
+            opts = model.simulate_options()
+            opts["sensitivities"] = ["d"]
+
+            res = model.simulate(options=opts)
+            nose.tools.assert_almost_equal(res.final('dx/dd'), 0.36789, 3)
+
+            assert res.solver.statistics["nsensfcnfcns"] > 0
+            assert res.solver.statistics["nfcnjacs"] == 0
+            
+        @testattr(stddist = True)
+        def test_basicsens2(self):
+            model = Dummy_FMUModelME2([], "BasicSens2.fmu", os.path.join(file_path, "files", "FMUs", "XML", "ME2.0"), _connect_dll=False)
+            
+            caps = model.get_capability_flags()
+            caps["providesDirectionalDerivatives"] = True
+            model.get_capability_flags = lambda : caps
+            
+            def f(*args, **kwargs):
+                d = model.values[model.variables["d"].value_reference]
+                x = model.continuous_states[0]
+                model.values[model.variables["der(x)"].value_reference] = d*x
+                return np.array([d*x])
+            
+            def d(*args, **kwargs):
+                if args[0][0] == 40:
+                    return np.array([-1.0])
+                else:
+                    return model.continuous_states
+                
+            model.get_directional_derivative = d
+            model.get_derivatives = f
+            model._provides_directional_derivatives = lambda : True
+            
+            opts = model.simulate_options()
+            opts["sensitivities"] = ["d"]
+
+            res = model.simulate(options=opts)
+            nose.tools.assert_almost_equal(res.final('dx/dd'), 0.36789, 3)
+            
+            assert res.solver.statistics["nsensfcnfcns"] == 0
+        
         @testattr(stddist = True)
         def test_relative_tolerance(self):
             model = Dummy_FMUModelME2([], "NoState.Example1.fmu", os.path.join(file_path, "files", "FMUs", "XML", "ME2.0"), _connect_dll=False)
@@ -361,8 +455,179 @@ if assimulo_installed:
             res = model.simulate(final_time=1.5,options=opts)
             
             assert res.solver.maxord == 1
+        
+        @testattr(stddist = True)
+        def test_with_jacobian_option(self):
+            model = Dummy_FMUModelME2([], "NoState.Example1.fmu", os.path.join(file_path, "files", "FMUs", "XML", "ME2.0"), _connect_dll=False)
+            opts = model.simulate_options()
+            opts["solver"] = "CVode"
+            opts["result_handling"] = "none"
+            
+            from pyfmi.fmi_algorithm_drivers import AssimuloFMIAlg, PYFMI_JACOBIAN_LIMIT
+            
+            class TempAlg(AssimuloFMIAlg):
+                def solve(self):
+                    pass
+            
+            def run_case(expected, default="Default"):
+                model.reset()
+                res = model.simulate(final_time=1.5,options=opts, algorithm=TempAlg)
+                assert res.options["with_jacobian"] == default, res.options["with_jacobian"]
+                assert res.solver.problem._with_jacobian == expected, res.solver.problem._with_jacobian
+            
+            run_case(False)
+            
+            model.get_ode_sizes = lambda: (PYFMI_JACOBIAN_LIMIT+1, 0)
+            run_case(True)
+            
+            opts["solver"] = "Radau5ODE"
+            run_case(False)
+            
+            opts["solver"] = "CVode"
+            opts["with_jacobian"] = False
+            run_case(False, False)
+            
+            model.get_ode_sizes = lambda: (PYFMI_JACOBIAN_LIMIT-1, 0)
+            opts["with_jacobian"] = True
+            run_case(True, True)
+        
+        @testattr(stddist = True)
+        def test_sparse_option(self):
+            from pyfmi.fmi_algorithm_drivers import AssimuloFMIAlg, PYFMI_JACOBIAN_SPARSE_SIZE_LIMIT, PYFMI_JACOBIAN_SPARSE_NNZ_LIMIT
+            
+            class TempAlg(AssimuloFMIAlg):
+                def solve(self):
+                    pass
+            
+            def run_case(expected_jacobian, expected_sparse, fnbr=0, nnz={}, set_sparse=False):
+                class Sparse_FMUModelME2(Dummy_FMUModelME2):
+                    def get_derivatives_dependencies(self):
+                        return (nnz, {})
+            
+                model = Sparse_FMUModelME2([], "NoState.Example1.fmu", os.path.join(file_path, "files", "FMUs", "XML", "ME2.0"), _connect_dll=False)
+                opts = model.simulate_options()
+                opts["solver"] = "CVode"
+                opts["result_handling"] = "none"
+                if set_sparse:
+                    opts["CVode_options"]["linear_solver"] = "SPARSE"
+                
+                model.get_ode_sizes = lambda: (fnbr, 0)
+                
+                res = model.simulate(final_time=1.5,options=opts, algorithm=TempAlg)
+                assert res.solver.problem._with_jacobian == expected_jacobian, res.solver.problem._with_jacobian
+                assert res.solver.linear_solver == expected_sparse, res.solver.linear_solver
+            
+            run_case(False, "DENSE")
+            run_case(True, "DENSE",  PYFMI_JACOBIAN_SPARSE_SIZE_LIMIT+1, {"Dep": [1]*PYFMI_JACOBIAN_SPARSE_SIZE_LIMIT**2})
+            run_case(True, "SPARSE", PYFMI_JACOBIAN_SPARSE_SIZE_LIMIT+1, {"Dep": [1]*PYFMI_JACOBIAN_SPARSE_SIZE_LIMIT})
+            run_case(True, "SPARSE", PYFMI_JACOBIAN_SPARSE_SIZE_LIMIT+1, {"Dep": [1]*PYFMI_JACOBIAN_SPARSE_SIZE_LIMIT}, True)
+        
+        @testattr(stddist = True)
+        def test_ncp_option(self):
+            model = Dummy_FMUModelME2([], "NoState.Example1.fmu", os.path.join(file_path, "files", "FMUs", "XML", "ME2.0"), _connect_dll=False)
+            opts = model.simulate_options()
+            assert opts["ncp"] == 500, opts["ncp"]
+        
+        @testattr(stddist = True)
+        def test_maxh_option(self):
+            model = Dummy_FMUModelME2([], "NoState.Example1.fmu", os.path.join(file_path, "files", "FMUs", "XML", "ME2.0"), _connect_dll=False)
+            opts = model.simulate_options()
+            opts["result_handling"] = "none"
+            
+            from pyfmi.fmi_algorithm_drivers import AssimuloFMIAlg
+            
+            class TempAlg(AssimuloFMIAlg):
+                def solve(self):
+                    pass
+            
+            def run_case(tstart, tstop, solver, ncp="Default"):
+                model.reset()
+                
+                opts["solver"] = solver
+                
+                if ncp != "Default":
+                    opts["ncp"] = ncp
+                
+                if opts["ncp"] == 0:
+                    expected = 0.0
+                else:
+                    expected = (float(tstop)-float(tstart))/float(opts["ncp"])
+                
+                res = model.simulate(start_time=tstart, final_time=tstop,options=opts, algorithm=TempAlg)
+                assert res.solver.maxh == expected, res.solver.maxh
+                assert res.options[solver+"_options"]["maxh"] == "Default", res.options[solver+"_options"]["maxh"]
+            
+            run_case(0,1,"CVode")
+            run_case(0,1,"CVode", 0)
+            run_case(0,1,"Radau5ODE")
+            run_case(0,1,"Dopri5")
+            run_case(0,1,"RodasODE")
+            run_case(0,1,"LSODAR")
+            run_case(0,1,"LSODAR")
+
             
 class Test_FMUModelME2:
+    
+    @testattr(stddist = True)
+    def test_estimate_directional_derivatives_linearstate(self):
+        full_path = os.path.join(file_path, "files", "FMUs", "XML", "ME2.0", "LinearStateSpace.fmu")
+        model = Dummy_FMUModelME2([], full_path, _connect_dll=False)
+        
+        def f(*args, **kwargs):
+            derx1 = -1.*model.values[model.variables["x[1]"].value_reference] + model.values[model.variables["u[1]"].value_reference]
+            derx2 = -1.*model.values[model.variables["x[2]"].value_reference] + model.values[model.variables["u[1]"].value_reference]
+            
+            model.values[model.variables["y[1]"].value_reference] = model.values[model.variables["x[1]"].value_reference] + model.values[model.variables["x[2]"].value_reference]
+
+            return np.array([derx1, derx2])
+        model.get_derivatives = f
+        
+        model.initialize()
+        model.event_update()
+        model.enter_continuous_time_mode()
+        
+        [As, Bs, Cs, Ds] = model.get_state_space_representation(use_structure_info=False)
+        [A, B, C, D] = model.get_state_space_representation()
+        
+        assert As.shape == A.shape, str(As.shape)+' '+str(A.shape)
+        assert Bs.shape == B.shape, str(Bs.shape)+' '+str(B.shape)
+        assert Cs.shape == C.shape, str(Cs.shape)+' '+str(C.shape)
+        assert Ds.shape == D.shape, str(Ds.shape)+' '+str(D.shape)
+        
+        assert np.allclose(As, A.toarray()), str(As)+' '+str(A.toarray())
+        assert np.allclose(Bs, B.toarray()), str(Bs)+' '+str(B.toarray())
+        assert np.allclose(Cs, C.toarray()), str(Cs)+' '+str(C.toarray())
+        assert np.allclose(Ds, D.toarray()), str(Ds)+' '+str(D.toarray())
+        
+    @testattr(stddist = True)
+    def test_estimate_directional_derivatives_without_structure_info(self):
+        full_path = os.path.join(file_path, "files", "FMUs", "XML", "ME2.0", "Bouncing_Ball.fmu")
+        model = Dummy_FMUModelME2([], full_path, _connect_dll=False)
+        
+        def f(*args, **kwargs):
+            derh = model.values[model.variables["v"].value_reference]
+            derv = -9.81
+            model.values[model.variables["der(h)"].value_reference] = derh
+            return np.array([derh, derv])
+        model.get_derivatives = f
+        
+        model.initialize()
+        model.event_update()
+        model.enter_continuous_time_mode()
+        
+        [As, Bs, Cs, Ds] = model.get_state_space_representation(use_structure_info=False)
+        [A, B, C, D] = model.get_state_space_representation()
+        
+        assert As.shape == A.shape, str(As.shape)+' '+str(A.shape)
+        assert Bs.shape == B.shape, str(Bs.shape)+' '+str(B.shape)
+        assert Cs.shape == C.shape, str(Cs.shape)+' '+str(C.shape)
+        assert Ds.shape == D.shape, str(Ds.shape)+' '+str(D.shape)
+        
+        assert np.allclose(As, A.toarray()), str(As)+' '+str(A.toarray())
+        assert np.allclose(Bs, B.toarray()), str(Bs)+' '+str(B.toarray())
+        assert np.allclose(Cs, C.toarray()), str(Cs)+' '+str(C.toarray())
+        assert np.allclose(Ds, D.toarray()), str(Ds)+' '+str(D.toarray())
+    
     @testattr(stddist = True)
     def test_estimate_directional_derivatives_BCD(self):
         full_path = os.path.join(file_path, "files", "FMUs", "XML", "ME2.0", "OutputTest2.fmu")
@@ -389,34 +654,33 @@ class Test_FMUModelME2:
             A = func(use_structure_info=True)
             B = func(use_structure_info=True, output_matrix=A)
             assert A is B #Test that the returned matrix is actually the same as the input
-            np.allclose(A.toarray(),B.toarray())
+            assert np.allclose(A.toarray(),B.toarray())
             A = func(use_structure_info=False)
             B = func(use_structure_info=False, output_matrix=A)
-            print(A)
             assert A is B
-            np.allclose(A,B)
+            assert np.allclose(A,B)
             C = func(use_structure_info=True, output_matrix=A)
             assert A is not C
-            np.allclose(C.toarray(), A)
+            assert np.allclose(C.toarray(), A)
             D = func(use_structure_info=False, output_matrix=C)
             assert D is not C
-            np.allclose(D, C.toarray())
+            assert np.allclose(D, C.toarray())
         
         B = model._get_B(use_structure_info=True)
         C = model._get_C(use_structure_info=True)
         D = model._get_D(use_structure_info=True)
         
-        np.allclose(B.toarray(), np.array([[0.0],[0.0]]))
-        np.allclose(C.toarray(), np.array([[0.0, 0.0],[0.0, 1.0], [1.0, 0.0]]))
-        np.allclose(D.toarray(), np.array([[-1.0],[0.0], [1.0]]))
+        assert np.allclose(B.toarray(), np.array([[0.0],[0.0]]))
+        assert np.allclose(C.toarray(), np.array([[0.0, 0.0],[0.0, 1.0], [1.0, 0.0]]))
+        assert np.allclose(D.toarray(), np.array([[-1.0],[0.0], [1.0]]))
         
         B = model._get_B(use_structure_info=False)
         C = model._get_C(use_structure_info=False)
         D = model._get_D(use_structure_info=False)
         
-        np.allclose(B, np.array([[0.0],[0.0]]))
-        np.allclose(C, np.array([[0.0, 0.0],[0.0, 1.0], [1.0, 0.0]]))
-        np.allclose(D, np.array([[-1.0],[0.0], [1.0]]))
+        assert np.allclose(B, np.array([[0.0],[0.0]]))
+        assert np.allclose(C, np.array([[0.0, 0.0],[0.0, 1.0], [1.0, 0.0]]))
+        assert np.allclose(D, np.array([[-1.0],[0.0], [1.0]]))
         
     @testattr(stddist = True)
     def test_output_dependencies(self):
@@ -441,8 +705,8 @@ class Test_FMUModelME2:
         
         [state_dep, input_dep] = model.get_output_dependencies()
         
-        assert len(state_dep.keys()) == 0
-        assert len(input_dep.keys()) == 0
+        assert len(state_dep.keys()) == 0, len(state_dep.keys())
+        assert len(input_dep.keys()) == 0, len(input_dep.keys())
     
     @testattr(stddist = True)
     def test_derivative_dependencies(self):
@@ -452,8 +716,8 @@ class Test_FMUModelME2:
         
         [state_dep, input_dep] = model.get_derivatives_dependencies()
         
-        assert len(state_dep.keys()) == 0
-        assert len(input_dep.keys()) == 0
+        assert len(state_dep.keys()) == 0, len(state_dep.keys())
+        assert len(input_dep.keys()) == 0, len(input_dep.keys())
     
     @testattr(stddist = True)
     def test_malformed_xml(self):
@@ -472,8 +736,8 @@ class Test_FMUModelME2:
     def test_units(self):
         model = FMUModelME2("CoupledClutches.fmu", os.path.join(file_path, "files", "FMUs", "XML", "ME2.0"), _connect_dll=False)
         
-        assert model.get_variable_unit("J1.w") == "rad/s"
-        assert model.get_variable_unit("J1.phi") == "rad"
+        assert model.get_variable_unit("J1.w") == "rad/s", model.get_variable_unit("J1.w")
+        assert model.get_variable_unit("J1.phi") == "rad", model.get_variable_unit("J1.phi")
         
         nose.tools.assert_raises(FMUException, model.get_variable_unit, "clutch1.useHeatPort")
         nose.tools.assert_raises(FMUException, model.get_variable_unit, "clutch1.sss")
@@ -483,19 +747,27 @@ class Test_FMUModelME2:
     def test_display_units(self):
         model = FMUModelME2("CoupledClutches.fmu", os.path.join(file_path, "files", "FMUs", "XML", "ME2.0"), _connect_dll=False)
         
-        assert model.get_variable_display_unit("J1.phi") == "deg"
+        assert model.get_variable_display_unit("J1.phi") == "deg", model.get_variable_display_unit("J1.phi")
         nose.tools.assert_raises(FMUException, model.get_variable_display_unit, "J1.w")
 
 class Test_FMUModelBase2:
+    @testattr(stddist = True)
+    def test_unicode_description(self):
+        full_path = os.path.join(file_path, "files", "FMUs", "XML", "ME2.0", "Description.fmu")
+        model = FMUModelME2(full_path, _connect_dll=False)
+        
+        desc = model.get_variable_description("x")
+
+        assert desc == "Test symbols '' ‘’"
     
     @testattr(stddist = True)
     def test_declared_enumeration_type(self):
         model = FMUModelME2("Enumerations.Enumeration3.fmu", os.path.join(file_path, "files", "FMUs", "XML", "ME2.0"), _connect_dll=False)
         
         enum = model.get_variable_declared_type("x")
-        assert len(enum.items.keys()) == 2
+        assert len(enum.items.keys()) == 2, len(enum.items.keys())
         enum = model.get_variable_declared_type("home")
-        assert len(enum.items.keys()) == 4
+        assert len(enum.items.keys()) == 4, len(enum.items.keys())
         
         nose.tools.assert_raises(FMUException, model.get_variable_declared_type, "z")
 
@@ -533,10 +805,10 @@ class Test_FMUModelBase2:
         vars = model.get_variable_alias("J4.phi")
         for var in vars:
             [r,i,b] = model.get_model_time_varying_value_references(filter=var)
-            assert len(r) == 1
+            assert len(r) == 1, len(r)
         
         [r,i,b] = model.get_model_time_varying_value_references(filter=list(vars.keys()))
-        assert len(r) == 1
+        assert len(r) == 1, len(r)
     
     @testattr(stddist = True)
     def test_get_directional_derivative_capability(self):
@@ -602,11 +874,11 @@ class Test_FMUModelBase2:
 
         sc_x = negated_alias.get_scalar_variable("x")
         
-        assert sc_x.name == "x"
-        assert sc_x.value_reference >= 0
-        assert sc_x.type == fmi.FMI2_REAL
-        assert sc_x.variability == fmi.FMI2_CONTINUOUS
-        assert sc_x.causality == fmi.FMI2_LOCAL
+        assert sc_x.name == "x", sc_x.name
+        assert sc_x.value_reference >= 0, sc_x.value_reference
+        assert sc_x.type == fmi.FMI2_REAL, sc_x.type
+        assert sc_x.variability == fmi.FMI2_CONTINUOUS, sc_x.variability
+        assert sc_x.causality == fmi.FMI2_LOCAL, sc_x.causality
 
         nose.tools.assert_raises(FMUException, negated_alias.get_scalar_variable, "not_existing")
     
@@ -623,25 +895,25 @@ class Test_load_fmu_only_XML:
         
         model = FMUModelME1("CoupledClutches.fmu", os.path.join(file_path, "files", "FMUs", "XML", "ME1.0"), _connect_dll=False)
         
-        assert model.get_name() == "CoupledClutches"
+        assert model.get_name() == "CoupledClutches", model.get_name()
 
     @testattr(stddist = True)
     def test_loading_xml_cs1(self):
         
         model = FMUModelCS1("CoupledClutches.fmu", os.path.join(file_path, "files", "FMUs", "XML", "CS1.0"), _connect_dll=False)
         
-        assert model.get_name() == "CoupledClutches"
+        assert model.get_name() == "CoupledClutches", model.get_name()
         
     @testattr(stddist = True)
     def test_loading_xml_me2(self):
         
         model = FMUModelME2("CoupledClutches.fmu", os.path.join(file_path, "files", "FMUs", "XML", "ME2.0"), _connect_dll=False)
         
-        assert model.get_name() == "CoupledClutches"
+        assert model.get_name() == "CoupledClutches", model.get_name()
         
     @testattr(stddist = True)
     def test_loading_xml_cs2(self):
         
         model = FMUModelCS2("CoupledClutches.fmu", os.path.join(file_path, "files", "FMUs", "XML", "CS2.0"), _connect_dll=False)
         
-        assert model.get_name() == "CoupledClutches"
+        assert model.get_name() == "CoupledClutches", model.get_name()
